#!/usr/bin/perl -w
#
# Bill Adams (bill@evil.nwpacfica.net)
#
# 26 Apr  1998  baa  Created Script.
# 30 Juli 1988  Hans de Goede, removed malloc stuff, beware this is my first
#               perl coding attempt ever ;)
use strict;

#Some defines.
my $equals = ' = ';
my $back_slash = "\\";  #No single quotes because emacs choaks.
my $obj_tag = "OBJ_";   #What to put at the front of each object section
my $slash = "/";        #For generating paths.
my $nl = "\n";	     #New Line
my $tab = "\t";         #Tab as needed by make.
my $comment = '#';      #Start of a comment in a makefile.

#All of the sources that do not have osd stuff.
my $core_makefile = 'makefile.objs';  

print "Generating some makefile includes.\n";

#Save the top-level directory.  
if (-d 'src') {chdir ('src')}  #Go into the src directory.
my $pwd = `pwd`; chomp($pwd);

#Be sure that we made it into the top-level source directory.
if (not -f 'mame.h') {die "Error: No mame.h in $ENV{PWD}\n"}

open (CORE_MAKEFILE, '>..'.$slash.$core_makefile)
    or die "Could not open ..$slash$core_makefile for writing.\n$!\n";

#The first thing to do is to check the dos makefile for what objects 
#  they have included.
my %dos_objects = &dos_objects();

my $dir;

print "Getting a list of include files...$nl";
my (@include_dirs, %lcase_dirs);
push (@include_dirs, '.');
foreach $dir (<*>) {
    next if (not -d $dir or $dir eq '..');
    push (@include_dirs, $dir);
    
    #Make it so we can to case-insensitive searches.
    my $lcase_dir = $dir;
    $lcase_dir =~ tr/[A-Z]/[a-z]/;
    $lcase_dirs{$lcase_dir} = $dir;
} 

print "Building the core makefile ($core_makefile) and$nl";
print "Checking/Fixing all #include's$nl";

my (@object_dirs);
foreach my $dir (<*>) {
    next if (not -d $dir or $dir eq '.' or $dir eq '..');
    push (@object_dirs, $dir);
    next if ($dir eq 'msdos' or $dir eq 'unix'); #No osd stuff in our makefile.
    print "Working on $dir\n";
    print CORE_MAKEFILE $obj_tag.$dir.$equals.$back_slash.$nl.$tab;
    my $count = 0;
    foreach my $file (<$dir$slash*.[ch]>) {
	#This way of reading directories can cause problems if there are a
	#  very large number of files in the directory.  Should be replaced
	#  with an opendir/readdir call.
	next if not -f $file;  #Just to be sure.
	
	#Verify all of the includes have the correct case in the name. 
	print $tab."Checking/Fixing $file->";
	&check_includes ($file);
	if ($file =~ m/(.*)\.[cC]/) {
	    if (++$count == 3) {
		$count = 0;
		print CORE_MAKEFILE $back_slash.$nl.$tab;
	    }
	    #$file =~ m/(.*)\.[cC]/;
	    print "Adding to $core_makefile ->";
	    my $object = 'obj'.$slash.$1.'.o';
	    $object =~ tr/[A-Z]/[a-z]/;
	    if (not defined $dos_objects{$object}) {
		print "Warning: obj".$slash.$1.
		    ".o is not defined in the dos makefile...skip.$nl";
		next;
	    }
	    print CORE_MAKEFILE  'obj'.$slash.$1.'.o  ';
	}
	print "ok.\n";
    }
    print CORE_MAKEFILE $nl.$nl;
    chdir ($pwd) or die $!;
}

print CORE_MAKEFILE $obj_tag."main".$equals.$back_slash.$nl.$tab;
my $count = 0;
foreach my $file (<*.[ch]>) {
    print $tab."Checking/Fixing $file->";
    &check_includes ($file);
    if ($file =~ m/(.*)\.[cC]/) {
        if (++$count == 3) {
	    $count = 0;
	    print CORE_MAKEFILE $back_slash.$nl.$tab;
        }
        #$file =~ m/(.*)\.[cC]/;
        print "Adding to $core_makefile ->";
        my $object = 'obj'.$slash.$1.'.o';
        $object =~ tr/[A-Z]/[a-z]/;
        if (not defined $dos_objects{$object}) {
            print "Warning: obj".$slash.$1.
            ".o is not defined in the dos makefile...skip.$nl";
            next;
        }
        print CORE_MAKEFILE  'obj'.$slash.$1.'.o  ';
    }
    print "Ok$nl";
}
print CORE_MAKEFILE $nl,$nl;

print CORE_MAKEFILE 'OBJCORE'.$equals.$back_slash.$nl.$tab;
my $count = 0;
foreach $dir ('main', @object_dirs) {
    if (++$count == 3) {
	$count = 0;
	print CORE_MAKEFILE $back_slash.$nl.$tab;
    }
    print CORE_MAKEFILE '$(OBJ_'.$dir.') ';
} print CORE_MAKEFILE $nl,$nl;

print CORE_MAKEFILE join 
    ($nl,
     $comment x 60,
     $comment.' These are the object subdirectories that need to be created.',
     '');

print CORE_MAKEFILE 'OBJDIRS'.$equals.'obj '.$back_slash.$nl.$tab;
$count = 0;
foreach $dir (@object_dirs) {
    if (++$count == 3) {
	$count = 0;
	print CORE_MAKEFILE $back_slash.$nl.$tab;
    }
    print CORE_MAKEFILE 'obj'.$slash.$dir.' ';
} print CORE_MAKEFILE $nl,$nl;



print CORE_MAKEFILE 'INCDIRS'.$equals.'-Isrc '.$back_slash.$nl.$tab;
$count = 0;
foreach $dir ('unix', @object_dirs) {
    if (++$count == 3) {
	$count = 0;
	print CORE_MAKEFILE $back_slash.$nl.$tab;
    }
    print CORE_MAKEFILE '-Isrc'.$slash.$dir.' ';
} print CORE_MAKEFILE $nl,$nl;



close (CORE_MAKEFILE);

print "Done Checking/Fixing\n";
exit;


sub check_includes {
    my ($old_file) = shift;
    my $tmp_file = $old_file.'.tmp';
    open (IN, $old_file) 
	or die $nl.$tab."Could not open [$old_file] for reading.$nl$!";
    open (OUT, '>'.$tmp_file) 
	or die $nl.$tab."Could not open [$tmp_file] for writing.$nl$!";
    #print "  Checking the includes of $old_file\n";
    my $line_number = 0;
    while (<IN>) {
	$line_number++;
	s/\r//g;
	if (m/(^\#include\s*\")(.*)\"/) {
	    #print "  Check include $2->";
	    if (-f $2) {
		print OUT; next;
	    } else {
		my $old_include = $_;
		chop($old_include);
		my $flag = 0;
		my $include_text = $1;
		my $include_file = $2;
		#print "Resolve->";
		my $orig_file = $2;
		#Check if there is a directory specified in the include.
		if ($2 =~ m/(.*)\/(.*)/) {
		    my $dir = $1;
		    my $file = $2;
		    if (not -d $dir) {
			my $lcase_dir = $dir; $lcase_dir =~ tr/[A-Z]/[a-z]/;
			if (not $lcase_dirs{$lcase_dir}) {
			    print "Warning: Could not resolve ".
				$include_file.$nl;
			    exit;
			}
			$dir = $lcase_dirs{$lcase_dir};
		    }
		    $file = &find_file($dir, $file);
		    if ($file) {
			my $new_include = $include_text.$dir.$slash.$file.'"';
			print '  '.$old_file.": ".$old_include.
			    " should be ".$nl.$tab.$new_include.$nl;
			print OUT $new_include.$nl;
			next;
		    } else {
			print "$old_file->Error: could not find $orig_file$nl";
			exit;
			next;
		    }
		} else {
		    # Search all of the dirs for the file.	
		    my $file;
		    foreach my $dir (@include_dirs) {
			if (-f $dir.$slash.$include_file) {
			    #We found it without looking case-insensitvly.
			    $flag = 1;
			    print OUT;
			    last;
			} elsif (($file = &find_file($dir, $include_file))) {
			    #We found it with a different case.
			    my $new_include = $include_text.$file.'"';
			    $flag = 1;
			    print '  '.$old_file.": ".$old_include.
				" should be ".$nl.$tab.$new_include.$nl;
			    print OUT $new_include.$nl; 
			    last;
			}
			
		    }
		}
		if (not $flag) { 
		    print "  $old_file->Warn: $include_file does not exist.\n";
		    sleep (2);
		} else {
		    #print OUT;
		}
	    }
	} else {
	    print OUT;
	}
    }
    close IN;
    close OUT;
    rename $tmp_file, $old_file 
	or die "Error: Could not rename $tmp_file to $old_file$nl$!";
    #sleep (1);
}

sub find_file {
    #Find a file in a give directory without regard to case.
    my $dir = shift;
    my $file = shift;
    $file =~ tr/[A-Z]/[a-z]/;  #Make the source file lowercase.
    foreach my $f (<$dir$slash*>) {
	$f =~ s/$dir$slash//;
	my $tmp_f = $f;           #Save the original file name.
	$f =~ tr/[A-Z]/[a-z]/;
	#print "  Compare $f to $file\n";
	if ($f eq $file) {
	    #print "Found $f in $dir [$tmp_f]$nl";
	    return $tmp_f;
	}
    }
    return '';
}

sub dos_objects {
    #Get a list of the dos objects from the dos makefile.  The indexes
    #  of the hash are all lowercase for easier matching.
    my %dos_objects;
    my $flag = 1;
    open (DOS_MAKEFILE, '..'.$slash.'makefile')
	or die "Could not open ..".$slash."makefile for reading.\n$!\n";
    print "Getting a list of objects from the dos makefile...$nl";
    while (<DOS_MAKEFILE>) {
	s/\r\n//;
	if (m/OBJS\s*\=(.*)/) {
	    $flag = 0;
	    $_ = $1;
	}
	next if $flag;
	m/^\s*(.*)\s*$/;
	$_ = $1;
	if (m/\\$/) {chop} else {$flag = 1}
	s/\s*$//;
	foreach my $object (split (/\s+/)) {
	    $object =~ tr/[A-Z]/[a-z]/;
	    $dos_objects{$object}++;
	}
    }
    return (%dos_objects);
}
